from zExceptions import Forbidden
from ZPublisher.HTTPRequest import HTTPRequest


def check(request):
    if isinstance(request, HTTPRequest):
        if request.get("REQUEST_METHOD", "GET").upper() != "POST":
            raise Forbidden("Request must be POST")



from plone.keyring.interfaces import IKeyManager
from plone.protect.interfaces import IDisableCSRFProtection
from plone.protect.utils import getRoot
from plone.protect.utils import getRootKeyManager
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent
from zope.component import adapter
from zope.component import ComponentLookupError
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from zope.interface import alsoProvides

import logging
import time


LOGGER = logging.getLogger("plone.protect")

_ring_rotation_schedules = (
    ("_system", 60 * 60 * 24 * 7),  # weekly
    ("_forms", 60 * 60 * 24),  # daily
    ("_anon", 60 * 60 * 24),  # daily
)


def _rotate(manager):
    current_time = time.time()
    for ring_name, check_period in _ring_rotation_schedules:
        try:
            ring = manager[ring_name]
            if (ring.last_rotation + check_period) < current_time:
                LOGGER.info("auto rotating keyring %s" % ring_name)
                ring.rotate()
        except KeyError:
            continue


@adapter(IUserLoggedInEvent)
def onUserLogsIn(event):
    """
    since we already write to the database when a user logs in,
    let's check for key rotation here
    """
    # disable csrf protection on login requests
    req = getRequest()
    if req:
        alsoProvides(req, IDisableCSRFProtection)

    try:
        manager = getUtility(IKeyManager)
        _rotate(manager)
        # also check rotation of zope root keyring
        root = getRoot(getSite())
        manager = getRootKeyManager(root)
        if manager:
            _rotate(manager)
    except ComponentLookupError:
        if req:
            url = req.URL
        else:
            url = ""
        LOGGER.warn("cannot find key manager for url %s" % url)



from plone.protect.postonly import check
from unittest import defaultTestLoader
from unittest import TestCase
from unittest import TestSuite
from zExceptions import Forbidden
from ZPublisher.HTTPRequest import HTTPRequest


class PostOnlyTests(TestCase):
    def makeRequest(self, method):
        return HTTPRequest(
            None,
            dict(REQUEST_METHOD=method, SERVER_PORT="80", SERVER_NAME="localhost"),
            None,
        )

    def testNonHTTPRequestAllowed(self):
        check("not a request")

    def testGETRequestForbidden(self):
        self.assertRaises(Forbidden, check, self.makeRequest("GET"))

    def testPOSTRequestAllowed(self):
        check(self.makeRequest("POST"))


def test_suite():
    return TestSuite((defaultTestLoader.loadTestsFromTestCase(PostOnlyTests),))



from AccessControl.requestmethod import buildfacade
from Acquisition import aq_parent
from BTrees.IFBTree import IFBTree
from BTrees.IIBTree import IIBTree
from BTrees.IOBTree import IOBTree
from BTrees.LFBTree import LFBTree
from BTrees.LLBTree import LLBTree
from BTrees.LOBTree import LOBTree
from BTrees.OIBTree import OIBTree
from BTrees.OLBTree import OLBTree
from BTrees.OOBTree import OOBTree
from OFS.interfaces import IApplication
from plone.keyring.keymanager import KeyManager
from plone.protect.authenticator import createToken
from zope.globalrequest import getRequest

import inspect
import logging


SAFE_WRITE_KEY = "plone.protect.safe_oids"
BTREE_TYPES = (
    IFBTree,
    IIBTree,
    IOBTree,
    LFBTree,
    LLBTree,
    LOBTree,
    OIBTree,
    OLBTree,
    OOBTree,
)
LOGGER = logging.getLogger("plone.protect")

_default = []

# This is based on AccessControl.requestmethod.postonly
# It should probably be updated to use the decorator module.


class protect:
    def __init__(self, *checkers):
        self.checkers = checkers

    def __call__(self, callable):
        try:
            spec = inspect.getfullargspec(callable)
        except AttributeError:
            # Python 2.7 compatibility
            spec = inspect.getargspec(callable)

        args = spec.args
        defaults = spec.defaults

        try:
            r_index = args.index("REQUEST")
        except ValueError:
            raise ValueError("No REQUEST parameter in callable signature")

        arglen = len(args)
        if defaults is not None:
            defaults = list(zip(args[arglen - len(defaults) :], defaults))
            arglen -= len(defaults)

        def _curried(*args, **kw):
            request = None

            if len(args) > r_index:
                request = args[r_index]

            for checker in self.checkers:
                checker(request)

            # Reconstruct keyword arguments
            if defaults is not None:
                args, kwparams = args[:arglen], args[arglen:]
                for positional, (key, default) in zip(kwparams, defaults):
                    if positional is _default:
                        kw[key] = default
                    else:
                        kw[key] = positional

            return callable(*args, **kw)

        # Build a facade, with a reference to our locally-scoped _curried
        facade_globs = dict(_curried=_curried, _default=_default)
        name = callable.__name__
        exec(buildfacade(name, callable, callable.__doc__), facade_globs)
        return facade_globs[name]


def addTokenToUrl(url, req=None, manager=None):
    if not url:
        return url
    if req is None:
        req = getRequest()
    if req is None or not url.startswith(req.SERVER_URL):
        # only transforms urls to same site
        return url
    if getattr(req, "environ", _default) is _default:
        # TestRequests have no environ.
        token = createToken(manager=manager)
    elif "_auth_token" not in req.environ:
        # Let's cache this value since this could be called
        # many times for one request.
        token = createToken(manager=manager)
        req.environ["_auth_token"] = token
    else:
        token = req.environ["_auth_token"]

    if "_authenticator" not in url:
        if "?" not in url:
            url += "?"
        else:
            url += "&"
        url += "_authenticator=" + token
    return url


def getRootKeyManager(root):
    if not IApplication.providedBy(root):
        return
    try:
        manager = root._key_manager
    except AttributeError:
        manager = root._key_manager = KeyManager()
        safeWrite(root._key_manager)
        safeWrite(root)
    return manager


def getRoot(context):
    while not IApplication.providedBy(context) and context is not None:
        context = aq_parent(context)
    return context


def safeWrite(obj, request=None):
    if request is None:
        request = getRequest()
    if request is None or getattr(request, "environ", _default) is _default:
        # Request not found or it is a TestRequest without an environment.
        LOGGER.debug("could not mark object as a safe write")
        return
    if SAFE_WRITE_KEY not in request.environ:
        request.environ[SAFE_WRITE_KEY] = []
    try:
        if obj._p_oid not in request.environ[SAFE_WRITE_KEY]:
            request.environ[SAFE_WRITE_KEY].append(obj._p_oid)
        if isinstance(obj, BTREE_TYPES):
            bucket = obj._firstbucket
            while bucket:
                if bucket._p_oid not in request.environ[SAFE_WRITE_KEY]:
                    request.environ[SAFE_WRITE_KEY].append(bucket._p_oid)
                bucket = bucket._next
    except AttributeError:
        LOGGER.debug("object you attempted to mark safe does not have an oid")



from plone.protect.testing import PROTECT_FUNCTIONAL_TESTING
from plone.protect.utils import addTokenToUrl
from plone.protect.utils import protect
from unittest import defaultTestLoader
from unittest import TestCase
from unittest import TestSuite

import unittest


def funcWithoutRequest():
    pass


def funcWithRequest(one, two, REQUEST=None):
    return (one, two)


class DummyChecker:
    def __call__(self, request):
        self.request = request


class DecoratorTests(TestCase):
    def testFunctionMustHaveRequest(self):
        protector = protect()
        self.assertRaises(ValueError, protector, funcWithoutRequest)

    def testArgumentsPassed(self):
        wrapped = protect()(funcWithRequest)
        self.assertEqual(wrapped("one", "two"), ("one", "two"))

    def testKeywordArguments(self):
        wrapped = protect()(funcWithRequest)
        self.assertEqual(wrapped(one="one", two="two"), ("one", "two"))

    def testMixedArguments(self):
        wrapped = protect()(funcWithRequest)
        self.assertEqual(wrapped("one", two="two"), ("one", "two"))

    def testRequestPassedToChecker(self):
        checker = DummyChecker()
        wrapped = protect(checker)(funcWithRequest)
        request = []
        wrapped("one", "two", request)
        self.assertTrue(checker.request is request)


class UrlTests(unittest.TestCase):
    layer = PROTECT_FUNCTIONAL_TESTING

    def testWithUrlFromSameDomain(self):
        url = addTokenToUrl("http://nohost/foobar", self.layer["request"])
        self.assertTrue("_authenticator=" in url)

    def testWithUrlFromOtherDomain(self):
        url = addTokenToUrl("http://other/foobar", self.layer["request"])
        self.assertTrue("_authenticator=" not in url)

    def testAddingWithQueryParams(self):
        url = addTokenToUrl("http://nohost/foobar?foo=bar", self.layer["request"])
        self.assertTrue("_authenticator=" in url)

    def testWithoutRequest(self):
        url = addTokenToUrl("http://nohost/foobar")
        self.assertTrue("_authenticator=" in url)

    def testWithNone(self):
        url = addTokenToUrl(None, self.layer["request"])
        self.assertTrue(not url)


def test_suite():
    return TestSuite(
        (
            defaultTestLoader.loadTestsFromTestCase(DecoratorTests),
            defaultTestLoader.loadTestsFromTestCase(UrlTests),
        )
    )



from zope.interface import Interface


class IAuthenticatorView(Interface):
    def token():
        """return token value"""

    def authenticator():
        """Return an xhtml snippet which sets an authenticator.

        This must be included inside a <form> element.
        """

    def verify():
        """
        Verify if the request contains a valid authenticator.
        """


class IDisableCSRFProtection(Interface):
    """
    Be able to disable on a per-request basis
    """


class IConfirmView(Interface):
    pass



from plone.protect.authenticator import check
from plone.protect.interfaces import IConfirmView
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from zExceptions import Forbidden
from zope.interface import implementer


@implementer(IConfirmView)
class ConfirmView(BrowserView):

    def __call__(self):
        check(self.request)
        urltool = getToolByName(self.context, "portal_url")
        original_url = getattr(self.request, "original_url", "")
        if not original_url or not urltool.isURLInPortal(original_url):
            raise Forbidden(f"url not in portal: {original_url}")
        return self.index()



from setuptools import find_packages
from setuptools import setup


version = "5.0.5"

setup(
    name="plone.protect",
    version=version,
    description="Security for browser forms",
    long_description="{}\n{}".format(
        open("README.rst").read(), open("CHANGES.rst").read()
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="zope security CSRF",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://github.com/plone/plone.protect",
    license="BSD",
    packages=find_packages(),
    namespace_packages=[
        "plone",
    ],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "collective.monkeypatcher",
        "lxml[cssselect]",
        "plone.keyring",
        "plone.scale",
        "plone.transformchain",
        "Products.CMFCore",
        "Products.GenericSetup",
        "Products.PluggableAuthService",
        "repoze.xmliter",
        "setuptools",
        "z3c.zcmlhook",
        "Zope",
    ],
    extras_require={
        "test": [
            "plone.app.testing",
            "plone.testing",
        ],
    },
)
